/*====================================+=====================================+
! File CServerList.cpp                ! Copyright (C) 2018-2020 Remi PASCAL !
+-------------------------------------+-------------------------------------+
! This file is part of kimfo.                                               !
! It is free software: you can redistribute it and/or modify it under the   !
! terms of the GNU General Public License as published by the Free Software !
! Foundation, either version 3 of the License, or any later version.        !
! This application is distributed in the hope that it will be useful, but   !
! WITHOUT ANY WARRANTY; without even the implied warranty of                !
! MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General !
! Public License for more details.                                          !
! You should have received a copy of the GNU General Public License along   !
! with this application. If not, see <http://www.gnu.org/licenses/>.        !
+==========================================================================*/



/*-------------------------------------------------------------------------*/
#include <wx/wx.h>
#include <wx/renderer.h>
#include <wx/regex.h>
#include <wx/tokenzr.h>
#include <wx/timer.h>
#include "CApp.h"
#include "CFrame.h"
#include "common/sc_download.h"
#include "CServerList.h"
#include "CDlgPrefs.h"
#include "flags/canada_16x16.xpm"
#include "flags/france_16x16.xpm"
#include "flags/germany_16x16.xpm"
#include "flags/poland_16x16.xpm"
#include "flags/united_kingdom_16x16.xpm"
/*-------------------------------------------------------------------------*/



/*-------------------------------------------------------------------------*/
void CServer::restore_data( const t_vec_server &vec_server_save )
{
   /*----------------------------------------------------------------------*/
   auto it_save = std::find_if( vec_server_save.begin(),
                                vec_server_save.end(),
                                [ this ]( const CServer &server_save )
                                {  return server_save.m_s_ref == m_s_ref ; }
                              ) ;
   if( it_save != vec_server_save.end() )
   {  /*-------------------------------------------------------------------*/
      m_boo_select = it_save->m_boo_select ;
      m_avail = it_save->m_avail ;
      /*-------------------------------------------------------------------*/
   }
   /*----------------------------------------------------------------------*/
}

/*-------------------------------------------------------------------------*/
wxString CServer::get_avail_url() const
{
   /*----------------------------------------------------------------------*/
   const wxString co_s_url_fmt
                = "https://www.ovh.com/engine/api/dedicated/server/"
                  "availabilities?country=%s&hardware=%s" ;
   /*----------------------------------------------------------------------*/
   return wxString::Format( co_s_url_fmt,
                            wxGetApp().get_region_code(), m_s_ref
                          ) ;
   /*----------------------------------------------------------------------*/
}

/*-------------------------------------------------------------------------*/
CServerList::CServerList( wxWindow *parent,
                          e_server_type sty, const char *p_c_main_url
                        )
           : wxListCtrl( parent, wxID_ANY, wxDefaultPosition,
                         wxDefaultSize, wxLC_REPORT | wxLC_VIRTUAL
                       ),
             m_sty( sty ), m_s_main_url( p_c_main_url )
{
   /*----------------------------------------------------------------------*/
   m_s_name = get_name() ;
   /*--( Data "available" widths )-----------------------------------------*/
   AppendColumn( _( "Name" ), wxLIST_FORMAT_LEFT, 140 ) ;
   AppendColumn( _( "Reference" ), wxLIST_FORMAT_LEFT, 130 ) ;
   AppendColumn( _( "Data center" ), wxLIST_FORMAT_LEFT, 130 ) ;
   AppendColumn( _( "Price" ), wxLIST_FORMAT_LEFT, 210 ) ;
   AppendColumn( _( "CPU" ), wxLIST_FORMAT_LEFT, 310 ) ;
   AppendColumn( _( "Memory" ), wxLIST_FORMAT_LEFT, 220 ) ;
   AppendColumn( _( "Storage" ), wxLIST_FORMAT_LEFT, 160 ) ;
   AppendColumn( _( "Network" ), wxLIST_FORMAT_LEFT, 120 ) ;
   /*----------------------------------------------------------------------*/
   EnableAlternateRowColours( true ) ;
   EnableCheckBoxes( true ) ;
   /*----------------------------------------------------------------------*/
   SetImageList( wxGetApp().m_frame->m_notebook->m_il, wxIMAGE_LIST_SMALL ) ;
   /*----------------------------------------------------------------------*/
   Bind( wxEVT_LIST_ITEM_ACTIVATED,
         [ this ]( wxListEvent &event )
         {  /*-------------------------------------------------------------*/
            const wxString s_url = get_order_url( event.GetIndex() ) ;
            if( !s_url.empty() ) { wxLaunchDefaultBrowser( s_url ) ; }
            /*-------------------------------------------------------------*/
         }
       ) ;
   Bind( wxEVT_LIST_ITEM_RIGHT_CLICK,
         [ this ]( wxListEvent &event )
         {  context_menu() ; }
       ) ;
   Bind( wxEVT_LIST_ITEM_CHECKED,
         [ this ]( wxListEvent &event )
         {  set_select( event.GetIndex(), true ) ; }
       ) ;
   Bind( wxEVT_LIST_ITEM_UNCHECKED,
         [ this ]( wxListEvent &event )
         {  set_select( event.GetIndex(), false ) ; }
       ) ;
   /*----------------------------------------------------------------------*/
   Bind( wxEVT_KEY_DOWN, &CServerList::OnKeyDown, this ) ;
   /*----------------------------------------------------------------------*/
}

/*-------------------------------------------------------------------------*/
wxItemAttr *CServerList::OnGetItemAttr( long l_item ) const
{
   /*----------------------------------------------------------------------*/
   wxItemAttr *p_lia = wxListCtrl::OnGetItemAttr( l_item ) ;
   const CServer *p_server = get_server( l_item ) ;
   /*----------------------------------------------------------------------*/
   if( p_server != nullptr && p_server->m_avail == CServer::AVA_OK )
   {
      /*-------------------------------------------------------------------*/
      static wxItemAttr st_attr( wxGetApp().m_frame->m_co_coul_avail_ok,
                                 p_lia == nullptr
                                 ? wxNullColour
                                 : p_lia->GetBackgroundColour(),
                                 GetFont().Bold()
                               ) ;
      /*-------------------------------------------------------------------*/
      p_lia = &st_attr ;
      /*-------------------------------------------------------------------*/
   }
   /*----------------------------------------------------------------------*/
   return p_lia ;
   /*----------------------------------------------------------------------*/
}

/*-------------------------------------------------------------------------*/
int CServerList::OnGetItemColumnImage( long l_item, long l_column ) const
{
   /*----------------------------------------------------------------------*/
   const int co_i_no_image = -1 ;
   /*----------------------------------------------------------------------*/
   if( l_column != COL_DC ) { return co_i_no_image ; }
   /*----------------------------------------------------------------------*/
   const CServer *p_server = get_server( l_item ) ;
   if( p_server == nullptr ) { return co_i_no_image ; }
   /*----------------------------------------------------------------------*/
   return p_server->m_i_img_dc ;
   /*----------------------------------------------------------------------*/
}

/*-------------------------------------------------------------------------*/
wxString CServerList::OnGetItemText( long l_item, long l_column ) const
{
   /*----------------------------------------------------------------------*/
   const CServer *p_server = get_server( l_item ) ;
   /*----------------------------------------------------------------------*/
   if( p_server == nullptr ) { return wxEmptyString ; }
   /*----------------------------------------------------------------------*/
   switch( l_column )
   {
      /*-------------------------------------------------------------------*/
      case COL_NAME    : return p_server->m_s_name ;
      case COL_REF     : return p_server->m_s_ref ;
      case COL_DC      : return p_server->m_s_dc ;
      case COL_PRICE   : return p_server->m_s_price ;
      case COL_CPU     : return p_server->m_s_cpu ;
      case COL_MEMORY  : return p_server->m_s_memory ;
      case COL_STORAGE : return p_server->m_s_storage ;
      case COL_NETWORK : return p_server->m_s_network ;
      /*-------------------------------------------------------------------*/
   }
   /*----------------------------------------------------------------------*/
   return wxEmptyString ;
   /*----------------------------------------------------------------------*/
}

/*-------------------------------------------------------------------------*/
bool CServerList::OnGetItemIsChecked( long l_item ) const
{
   /*----------------------------------------------------------------------*/
   const CServer *p_server = get_server( l_item ) ;
   /*----------------------------------------------------------------------*/
   return p_server != nullptr && p_server->m_boo_select ;
   /*----------------------------------------------------------------------*/
}

/*-------------------------------------------------------------------------*/
void CServerList::OnKeyDown( wxKeyEvent &event )
{
   /*----------------------------------------------------------------------*/
   bool boo_cmd_key
      = ( ( event.GetModifiers() & ~wxMOD_ALT ) == wxMOD_CMD ) ;
   bool boo_alt_key
      = ( ( event.GetModifiers() & wxMOD_ALT ) == wxMOD_ALT ) ;
   /*----------------------------------------------------------------------*/
   if( !boo_alt_key && !boo_cmd_key && event.GetKeyCode() == ' ' )
   {  select_unselect_hil_following_focus() ; }
   else
   {  event.Skip() ; }
   /*----------------------------------------------------------------------*/
}

/*-------------------------------------------------------------------------*/
int CServerList::set_select( long l_item, bool boo_select )
{
   /*----------------------------------------------------------------------*/
   if( l_item < 0 || l_item >= static_cast< int >( m_vec_server.size() ) )
   {  return -1 ; }
   /*----------------------------------------------------------------------*/
   m_vec_server[ l_item ].m_boo_select = boo_select ;
   RefreshItem( l_item ) ;
   /*----------------------------------------------------------------------*/
   return 0 ;
   /*----------------------------------------------------------------------*/
}

/*-------------------------------------------------------------------------*/
void CServerList::refresh_all_items()
{
   /*----------------------------------------------------------------------*/
   const int co_i_last = GetItemCount() - 1 ;
   /*----------------------------------------------------------------------*/
   if( co_i_last >= 0 ) { RefreshItems( 0, co_i_last ) ; }
   /*----------------------------------------------------------------------*/
}

/*-------------------------------------------------------------------------*/
void CServerList::select_all( bool boo_select )
{
   /*----------------------------------------------------------------------*/
   for( CServer &server : m_vec_server )
   {  server.m_boo_select = boo_select ; }
   /*----------------------------------------------------------------------*/
   refresh_all_items() ;
   /*----------------------------------------------------------------------*/
}

/*-------------------------------------------------------------------------*/
void CServerList::select_invert()
{
   /*----------------------------------------------------------------------*/
   for( CServer &server : m_vec_server )
   {  server.m_boo_select = !server.m_boo_select ; }
   /*----------------------------------------------------------------------*/
   refresh_all_items() ;
   /*----------------------------------------------------------------------*/
}

/*-------------------------------------------------------------------------*/
void CServerList::select_unselect_hil_following_focus()
{
   /*----------------------------------------------------------------------*/
   const CServer *p_server = get_server( get_item_focus() ) ;
   if( p_server == nullptr ) { return ; }
   /*----------------------------------------------------------------------*/
   const bool co_boo_select = !p_server->m_boo_select ;
   /*----------------------------------------------------------------------*/
   for( long l_item = -1 ; ( l_item = get_next_item_hil( l_item ) ) != -1 ; )
   {  set_select( l_item, co_boo_select ) ; }
   /*----------------------------------------------------------------------*/
}

/*-------------------------------------------------------------------------*/
const CServer *CServerList::get_server( long l_item ) const
{
   /*----------------------------------------------------------------------*/
   if( l_item < 0 || l_item >= static_cast< int >( m_vec_server.size() ) )
   {  return nullptr ; }
   /*----------------------------------------------------------------------*/
   return &m_vec_server[ l_item ] ;
   /*----------------------------------------------------------------------*/
}

/*-------------------------------------------------------------------------*/
CServer *CServerList::get_server( long l_item )
{
   /*----------------------------------------------------------------------*/
   if( l_item < 0 || l_item >= static_cast< int >( m_vec_server.size() ) )
   {  return nullptr ; }
   /*----------------------------------------------------------------------*/
   return &m_vec_server[ l_item ] ;
   /*----------------------------------------------------------------------*/
}

/*-------------------------------------------------------------------------*/
CServer *CServerList::find_server( const wxString &s_ref )
{
   /*----------------------------------------------------------------------*/
   auto it = std::find_if( m_vec_server.begin(), m_vec_server.end(),
                           [ &s_ref ]( const CServer &server )
                           {  return server.m_s_ref == s_ref ; }
                         ) ;
   /*----------------------------------------------------------------------*/
   return it == m_vec_server.end() ? nullptr : &*it ;
   /*----------------------------------------------------------------------*/
}

/*-------------------------------------------------------------------------*/
wxString CServerList::get_order_url( long l_item ) const
{
   /*----------------------------------------------------------------------*/
   const CServer *p_server = get_server( l_item ) ;
   wxASSERT( p_server != nullptr ) ;
   /*--( The order url starts with a '/' )---------------------------------*/
   return m_s_main_url + p_server->m_s_order_url ;
   /*----------------------------------------------------------------------*/
}

/*-------------------------------------------------------------------------*/
wxString CServerList::get_detail_url( long l_item ) const
{
   /*----------------------------------------------------------------------*/
   const CServer *p_server = get_server( l_item ) ;
   wxASSERT( p_server != nullptr ) ;
   /*--( The detail url starts with a '/' )--------------------------------*/
   return m_s_main_url + p_server->m_s_detail_url ;
   /*----------------------------------------------------------------------*/
}

/*-------------------------------------------------------------------------*/
wxString CServerList::get_main_url() const
{
   /*----------------------------------------------------------------------*/
   return wxString::Format( "%s/%s", m_s_main_url,
                            wxGetApp().get_region_page( m_sty )
                          ) ;
   /*----------------------------------------------------------------------*/
}

/*--( Starting the download of the server web page )-----------------------*/
int CServerList::launch_refresh()
{
   /*----------------------------------------------------------------------*/
   CFrame *frame = wxGetApp().m_frame ;
   const wxString co_s_url
                = wxString::Format( "%s/%s",
                                    m_s_main_url,
                                    wxGetApp().get_region_page( m_sty )
                                  ) ;
   /*----------------------------------------------------------------------*/
   sc::CDownload dl( frame->m_notebook, co_s_url ) ;
   dl.get_dlbase().set_download_id( DID_SERVER_TYPE_START + m_sty ) ;
   dl.thread_start() ;
   /*----------------------------------------------------------------------*/
   return 0 ;
   /*----------------------------------------------------------------------*/
}

/*-------------------------------------------------------------------------*/
int CServerList::pre_refresh( int i_err_code, const wxString &co_s_data )
{
   /*----------------------------------------------------------------------*/
   CFrame *frame = wxGetApp().m_frame ;
   const wxString co_s_title = '\t' + m_s_name ;
   /*----------------------------------------------------------------------*/
   if( i_err_code != sc::CDownloadBase::DE_NO_ERROR_DONE )
   {  frame->log_error( wxString::Format( _( "%s: %s" ), co_s_title,
                              sc::CDownloadBase::st_get_err_msg( i_err_code )
                                          )
                        ) ;
      return -1 ;
   }
   /*----------------------------------------------------------------------*/
   if( co_s_data.empty() )
   {  frame->log_message( wxString::Format( _( "%s: no data" ), co_s_title
                                          )
                        ) ;
      return -2 ;
   }
   /*----------------------------------------------------------------------*/
   return 0 ;
   /*----------------------------------------------------------------------*/
}

/*-------------------------------------------------------------------------*/
int CServerList::post_refresh( int i_ret_pre )
{
   /*----------------------------------------------------------------------*/
   CFrame    *frame    = wxGetApp().m_frame ;
   CNotebook *notebook = frame->m_notebook ;

   /*----------------------------------------------------------------------*/
   for( auto &server : m_vec_server )
   {
      /*-------------------------------------------------------------------*/
      if( server.m_s_dc == "bhs" )
      {  server.m_i_img_dc = notebook->m_i_img_flag_canada ; }
      else
      if( server.m_s_dc == "gra rbx sbg" || server.m_s_dc == "fr" )
      {  server.m_i_img_dc = notebook->m_i_img_flag_france ; }
      else
      if( server.m_s_dc == "fra" )
      {  server.m_i_img_dc = notebook->m_i_img_flag_germany ; }
      else
      if( server.m_s_dc == "waw" )
      {  server.m_i_img_dc = notebook->m_i_img_flag_poland ; }
      else
      if( server.m_s_dc == "lon" )
      {  server.m_i_img_dc = notebook->m_i_img_flag_united_kingdom ; }
      /*-------------------------------------------------------------------*/
   }
   SetItemCount( m_vec_server.size() ) ;
   refresh_all_items() ;

   /*--( If everything went well the ok message is displayed )-------------*/
   if( i_ret_pre == 0 )
   {  /*--( wxPLURAL returns the second string if count == 0 ... )---------*/
      const int co_i_plural = ( m_vec_server.size() == 0 ? 1 : 2 ) ;
      /*-------------------------------------------------------------------*/
      wxString s_msg ;
      s_msg.Printf( '\t' + wxPLURAL( _( "%2zu %s server found" ),
                                     _( "%2zu %s servers found" ),
                                     co_i_plural
                                    ),
                    m_vec_server.size(), m_s_name
                  ) ;
      frame->log_message( s_msg ) ;
      /*-------------------------------------------------------------------*/
   }

   /*--( We have things to do once all sources have been refreshed )-------*/
   if( ++notebook->m_i_source_refresh_done == STY_NB )
   {
      /*--------------------------------------------------------------------+
      ! Actions that have to be done after the first refresh :              !
      ! - The selection of the servers that were selected at last           !
      !   application exit                                                  !
      +--------------------------------------------------------------------*/
      if( notebook->m_i_nb_refresh_server == 1 )
      {  /*----------------------------------------------------------------*/
         notebook->select_ref_to_select() ;
         /*----------------------------------------------------------------*/
      }
      /*-------------------------------------------------------------------*/
      frame->gui_enable_refresh_server( true ) ;
      frame->gui_enable_check_avail( true ) ;
      /*-------------------------------------------------------------------*/
   }
   /*----------------------------------------------------------------------*/
   return i_ret_pre ;
   /*----------------------------------------------------------------------*/
}

/*--( This function is supposed to be called once per server availability )*/
int CServerList::check_avail( int i_err_code,
                              const wxString &co_s_ref,
                              const wxString &co_s_data
                            )
{
   /*----------------------------------------------------------------------*/
   CFrame    *frame    = wxGetApp().m_frame ;
   CNotebook *notebook = frame->m_notebook  ;
   CServer   *p_server ;
   int       i_ret = 0 ;

   /*----------------------------------------------------------------------*/
   const wxString co_s_title = '\t' + m_s_name ;

   /*--( At this date, 404 is returned for some not available SYS servers )*/
   if( i_err_code == sc::CDownloadBase::DE_FILE_NOT_FOUND )
   {  frame->log_message( wxString::Format( _( "%s: %s NOT available" ),
                                            co_s_title, co_s_ref
                                          )
                        ) ;
      i_ret = -1 ; goto function_end ;
   }
   /*----------------------------------------------------------------------*/
   if( i_err_code != sc::CDownloadBase::DE_NO_ERROR_DONE )
   {  frame->log_message( wxString::Format( _( "%s: %s" ), co_s_title,
                              sc::CDownloadBase::st_get_err_msg( i_err_code )
                                          )
                        ) ;
      i_ret = -2 ; goto function_end ;
   }
   /*----------------------------------------------------------------------*/
   if( co_s_ref.empty() )
   {  frame->log_message( wxString::Format( _( "%s: no server reference !" ),
                                            co_s_title
                                          )
                        ) ;
      i_ret = -3 ; goto function_end ;
   }
   /*----------------------------------------------------------------------*/
   if( co_s_data.empty() )
   {  frame->log_message( wxString::Format( _( "%s: no data" ), co_s_title
                                          )
                        ) ;
      i_ret = -4 ; goto function_end ;
   }

   /*--( Known server reference ? )----------------------------------------*/
   if( ( p_server = find_server( co_s_ref ) ) == nullptr )
   {  frame->log_message( wxString::Format(
                                       _( "%s: reference \"%s\" unknown !" ),
                                       co_s_title, co_s_ref
                                          )
                        ) ;
      i_ret = -5 ; goto function_end ;
   }

   /*--( Is this server available ? )--------------------------------------*/
   if( sc::match( co_s_data, R"=("availability":"\d)=" ) <= 0 )
   {  /*-------------------------------------------------------------------*/
      p_server->m_avail = CServer::AVA_KO ;
      /*--------------------------------------------------------------------+
      ! The message is a bit different from the previous "NOT available"    !
      ! one to be able to recognize its origin                              !
      +--------------------------------------------------------------------*/
      frame->log_message( wxString::Format( _( "%s: %s NOT available." ),
                                            co_s_title, co_s_ref
                                          )
                        ) ;
      /*-------------------------------------------------------------------*/
   }
   else
   {  /*--( Signal this availability )-------------------------------------*/
      p_server->m_avail = CServer::AVA_OK ;
      /*-------------------------------------------------------------------*/
      frame->log_message_ok( wxString::Format( _( "%s: %s AVAILABLE" ),
                                               co_s_title, co_s_ref
                                             )
                           ) ;
      /*--( This can be done because of "std::vector" )--------------------*/
      const long l_item = p_server - get_server( 0 ) ;
      RefreshItem( l_item ) ;
      /*--( This item will be shown after complete execution )-------------*/
      if( notebook->m_i_page_to_show_avail_ok < 0 )
      {  notebook->m_i_page_to_show_avail_ok = m_sty ;
         notebook->m_l_item_to_show_avail_ok = l_item ;
      }
      /*-------------------------------------------------------------------*/
   }

   /*----------------------------------------------------------------------*/
function_end :
   /*--( Some work to do when all availabilities have been refreshed )-----*/
   if(    ++notebook->m_i_avail_refresh_done
       == notebook->m_i_avail_refresh_todo
     )
   {  /*-------------------------------------------------------------------*/
      frame->gui_enable_refresh_server( true ) ;
      frame->gui_enable_check_avail( true ) ;
      /*--( Alert user about the positive result ! )-----------------------*/
      if( notebook->m_i_page_to_show_avail_ok >= 0 )
      {  frame->alert_selected_server_avail() ; }
      /*-------------------------------------------------------------------*/
   }
   /*----------------------------------------------------------------------*/
   return i_ret ;
   /*----------------------------------------------------------------------*/
}

/*-------------------------------------------------------------------------*/
void CServerList::context_menu()
{
   /*--( Current server )--------------------------------------------------*/
   long l_item = get_item_focus() ;
   if( l_item < 0 || l_item >= GetItemCount() ) { return ; }
   /*--( Current server )--------------------------------------------------*/
   const CServer *p_server = get_server( l_item ) ;
   /*----------------------------------------------------------------------*/
   wxMenu menu ;
   int i_id = 0 ;
   const int co_i_id_open_order_url  = i_id++ ;
   const int co_i_id_copy_order_url  = i_id++ ;
   const int co_i_id_open_detail_url = i_id++ ;
   const int co_i_id_copy_detail_url = i_id++ ;
   const int co_i_id_copy_avail_url  = i_id   ;
   /*----------------------------------------------------------------------*/
   sc::append_item( &menu, co_i_id_open_order_url,
                    _( "Open order page in &browser" ),
                    sc::get_menu_bmp( wxART_EXECUTABLE_FILE )
                  ) ;
   sc::append_item( &menu, co_i_id_copy_order_url,
                    _( "Copy &order page address" ),
                    sc::get_menu_bmp( wxART_COPY )
                  ) ;
   /*----------------------------------------------------------------------*/
   if( !p_server->m_s_detail_url.empty() )
   {
      /*-------------------------------------------------------------------*/
      menu.AppendSeparator() ;
      sc::append_item( &menu, co_i_id_open_detail_url,
                       _( "Open detail page in b&rowser" ),
                       sc::get_menu_bmp( wxART_EXECUTABLE_FILE )
                     ) ;
      sc::append_item( &menu, co_i_id_copy_detail_url,
                       _( "Copy &detail page address" ),
                       sc::get_menu_bmp( wxART_COPY )
                     ) ;
      /*-------------------------------------------------------------------*/
   }
   /*----------------------------------------------------------------------*/
   menu.AppendSeparator() ;
   sc::append_item( &menu, co_i_id_copy_avail_url,
                    _( "Copy &availability address" ),
                    sc::get_menu_bmp( wxART_COPY )
                  ) ;
   /*----------------------------------------------------------------------*/
   int i_sel_id = GetPopupMenuSelectionFromUser( menu ) ;
   if( i_sel_id == co_i_id_open_order_url )
   {  wxLaunchDefaultBrowser( get_order_url( l_item ) ) ; }
   else
   if( i_sel_id == co_i_id_copy_order_url )
   {  sc::set_clipboard( get_order_url( l_item ) ) ; }
   else
   if( i_sel_id == co_i_id_open_detail_url )
   {  wxLaunchDefaultBrowser( get_detail_url( l_item ) ) ; }
   else
   if( i_sel_id == co_i_id_copy_detail_url )
   {  sc::set_clipboard( get_detail_url( l_item ) ) ; }
   else
   if( i_sel_id == co_i_id_copy_avail_url )
   {  sc::set_clipboard( get_server( l_item )->get_avail_url() ) ; }
   /*----------------------------------------------------------------------*/
}

/*-------------------------------------------------------------------------*/
int CSourceKimsufi::refresh( int i_err_code, const wxString &co_s_data )
{
   /*----------------------------------------------------------------------*/
   int i_ret_pre = pre_refresh( i_err_code, co_s_data ) ;
   if( i_ret_pre != 0 ) { return post_refresh( i_ret_pre ) ; }
   /*--( Save and reset previous server data )-----------------------------*/
   t_vec_server vec_server_save ;
   m_vec_server.swap( vec_server_save ) ;
   /*--( Replace "unbreakable space" characters )--------------------------*/
   wxString s_data = co_s_data ;
   s_data.Replace( wxString::FromUTF8( u8"\u00A0" ), " " ) ;
   /*----------------------------------------------------------------------*/
   const wxString co_s_re
       = R"=(\mclass="zone-dedicated-availability")="
         R"=(.*?\mdata-ref="(.*?)")="   // Reference
         R"=(.*?\m(KS-\d*)<)="          // Name
         R"=(.*?<td.*?>(.*?)</td>)="    // CPU Brand & Type
         R"=(.*?<td.*?>(.*?)</td>)="    // CPU Cores/Threads
         R"=(.*?<td.*?>(.*?)</td>)="    // CPU Speed
         R"=(.*?<td.*?>(.*?)</td>)="    // Memory
         R"=(.*?<td.*?>(.*?)</td>)="    // Storage
         R"=(.*?<td.*?>(.*?)</td>)="    // Net speed
         R"=(.*?<td.*?>(.*?)</td>)="    // Net mask
         R"=(.*?<td.*?>(.*?)</td>)="    // Price
         R"=(.*?\mdata-dc="(.*?)")="    // Data center
         R"=(.*?href="(.*?)")="         // Order url
         ;
   /*----------------------------------------------------------------------*/
   wxRegEx re( co_s_re, wxRE_ADVANCED ) ;
   wxRegEx re_clean_name( R"=(<span .*?</span>|<br>)=", wxRE_ADVANCED ) ;
   wxRegEx re_clean_space( R"=(\s+)=", wxRE_ADVANCED ) ;
   wxRegEx re_clean( R"=(<span[^>]*>|</span>|<br>|<del>.*?</del>)=",
                     wxRE_ADVANCED
                   ) ;
   /*----------------------------------------------------------------------*/
   const wxString co_s_col_info_sep
                = wxGetApp().m_frame->m_notebook->m_co_s_col_info_sep ;
   /*----------------------------------------------------------------------*/
   const wxChar *p_c_data = s_data.c_str() ;
   size_t sz_start ;
   size_t sz_len   ;
   for( ; re.Matches( p_c_data ) ; p_c_data += sz_start + sz_len )
   {
      /*--( Keep the found element to skip it later )----------------------*/
      re.GetMatch( &sz_start, &sz_len, 0 ) ;
      /*-------------------------------------------------------------------*/
      CServer server ;
      server.m_s_ref  = re.GetMatch( p_c_data, 1 ) ;
      /*--( Clean potential "flash sale" data )----------------------------*/
      server.m_s_name = re.GetMatch( p_c_data, 2 ) ;
      re_clean_name.ReplaceAll( &server.m_s_name, "" ) ;
      /*--( Extract the different parts of the server description )--------*/
      const size_t co_sz_nb_match = re.GetMatchCount() ;
      wxString s_val ;
      for( size_t sz_num = 3 ; sz_num < co_sz_nb_match - 2 ; ++sz_num )
      {
         /*--( Extracting and "cleaning" a value )-------------------------*/
         s_val = re.GetMatch( p_c_data, sz_num ) ;
         re_clean_space.ReplaceAll( &s_val, " " ) ;
         re_clean.ReplaceAll( &s_val, "" ) ;
         s_val = s_val.Trim().Trim( false ) ;
         /*----------------------------------------------------------------*/
         switch( sz_num )
         {
            /*-------------------------------------------------------------*/
            case  4 : case 5 :
               server.m_s_cpu += co_s_col_info_sep ; // Fallthrough
            case  3 :
               server.m_s_cpu += s_val ;
               break ;
            /*-------------------------------------------------------------*/
            case  6 : server.m_s_memory = s_val ; break ;
            case  7 : server.m_s_storage = s_val ; break ;
            /*-------------------------------------------------------------*/
            case  9 : server.m_s_network += ' '   ; // Fallthrough
            case  8 : server.m_s_network += s_val ;
               break ;
            /*-------------------------------------------------------------*/
            case 10 : server.m_s_price = s_val ; break ;
            /*-------------------------------------------------------------*/
         }
         /*----------------------------------------------------------------*/
      }

      /*--( The last elements which don't need any modification )----------*/
      server.m_s_dc        = re.GetMatch( p_c_data, 11 ) ;
      server.m_s_order_url = re.GetMatch( p_c_data, 12 ) ;

      /*--( Restore some data previously associated to this server )-------*/
      server.restore_data( vec_server_save ) ;

      /*--( Create the vector element )------------------------------------*/
      m_vec_server.emplace_back( server ) ;
      /*-------------------------------------------------------------------*/
   }

   /*----------------------------------------------------------------------*/
   return post_refresh( 0 ) ;
   /*----------------------------------------------------------------------*/
}

/*--( Server data are supposed to be set with the "group" values )---------*/
void CSourceSoYouStart::add_one_server_per_dc( CServer &server,
                                               const wxString &s_dc,
                                          const t_vec_server &vec_server_save
                                             )
{
   /*-----------------------------------------------------------------------+
   ! There is as many servers (ie references) as "country" options in       !
   ! the "dc" data. Therefore "dc data" is split and server records are     !
   ! created                                                                !
   +-----------------------------------------------------------------------*/
   wxRegEx re_dc( R"=(data-dc="(.*?)" data-plan-code="(.*?)")=",
                  wxRE_ADVANCED
                ) ;
   size_t sz_start_dc ;
   size_t sz_len_dc   ;
   wxString s_prev_serv_ref ;
   /*----------------------------------------------------------------------*/
   for( const wxChar *p_c_data_dc = s_dc.c_str() ;
        re_dc.Matches( p_c_data_dc ) ;
        p_c_data_dc += sz_start_dc + sz_len_dc
      )
   {
      /*--( Keep the found element to skip it later )----------------------*/
      re_dc.GetMatch( &sz_start_dc, &sz_len_dc, 0 ) ;
      /*-------------------------------------------------------------------*/
      server.m_s_dc  = re_dc.GetMatch( p_c_data_dc, 1 ) ;
      server.m_s_ref = re_dc.GetMatch( p_c_data_dc, 2 ) ;

      /*--------------------------------------------------------------------+
      ! Only one url of each kind is present using the first server         !
      ! reference of the dc "select" list. We keep the previous one to      !
      ! replace it with the new one.                                        !
      +--------------------------------------------------------------------*/
      if( !s_prev_serv_ref.empty() )
      {  server.m_s_order_url.Replace( s_prev_serv_ref, server.m_s_ref ) ;
         server.m_s_detail_url.Replace( s_prev_serv_ref, server.m_s_ref ) ;
      }
      s_prev_serv_ref = server.m_s_ref ;

      /*--( Restore some data previously associated to this server )-------*/
      server.restore_data( vec_server_save ) ;
      /*--( Create the vector element )------------------------------------*/
      m_vec_server.push_back( server ) ;
      /*-------------------------------------------------------------------*/
   }
   /*----------------------------------------------------------------------*/
}

/*-------------------------------------------------------------------------*/
void CSourceSoYouStart::refresh_deals( const wxString &co_s_data,
                                       const t_vec_server &vec_server_save
                                     )
{
   /*-----------------------------------------------------------------------+
   ! The So-You-Start server data with similar structures can be divided    !
   ! into two parts :                                                       !
   ! - The "advertised" offers : big individual squares, normally 3 of them !
   ! - The "listed" offers : One or more "table" per "country"              !
   +-----------------------------------------------------------------------*/
   const wxString co_s_re_serv
       = R"=(<p class="text--large weight--bold margin-bottom--20")="
         R"=(.*?>(.*?)<)="                              // Name
                                                        // Old price (unused)
         R"=(.*?(?:<del .*?</del></div>)?<span class="spanWithPrice red")="
         R"=(.*?<span class="red".*?>(.*?)</div><div)=" // Price
         R"=(.*?<select>(.*?)</select>)="               // Data Center
         R"=(.*?href="(.*?)")="                         // Order url
         R"=(.*?href="(.*?)")="                         // Info url
         R"=(.*?translate="none">(.*?)<br>)="           // CPU Brand & Type
         R"=((.*?)\()="                                 // CPU Cores/Threads
         R"=((.*?)\).*?<)="                             // CPU Speed
         R"=(.*?<div .*?>(.*?)</div>)="                 // Memory
         R"=(.*?<div .*?>(.*?)</div>)="                 // Storage
         R"=((?:<div .*?>[^:]+: (.*?)</div>)?</div>)="  // Network (possible)
         ;
   /*----------------------------------------------------------------------*/
   wxRegEx re_serv( co_s_re_serv, wxRE_ADVANCED ) ;
   wxRegEx re_clean_name( R"=(<span .*?</span>|<br>)=", wxRE_ADVANCED ) ;
   wxRegEx re_clean_space( R"=(\s+)=", wxRE_ADVANCED ) ;
   wxRegEx re_clean( R"=(<span[^>]*>|</span>|<br>)=", wxRE_ADVANCED ) ;
   /*----------------------------------------------------------------------*/
   const wxString co_s_col_info_sep
                = wxGetApp().m_frame->m_notebook->m_co_s_col_info_sep ;
   /*----------------------------------------------------------------------*/
   size_t sz_start_serv ;
   size_t sz_len_serv   ;
   /*----------------------------------------------------------------------*/
   for( const wxChar *p_c_data_serv = co_s_data.c_str() ;
        re_serv.Matches( p_c_data_serv ) ;
        p_c_data_serv += sz_start_serv + sz_len_serv
      )
   {
      /*--( Keep the found element to skip it later )----------------------*/
      re_serv.GetMatch( &sz_start_serv, &sz_len_serv, 0 ) ;
      /*-------------------------------------------------------------------*/
      CServer server ;
      /*--( Clean potential "flash sale" data )----------------------------*/
      server.m_s_name = re_serv.GetMatch( p_c_data_serv, 1 ) ;
      re_clean_name.ReplaceAll( &server.m_s_name, "" ) ;
      /*--( Extract the different parts of the server description )--------*/
      const size_t co_sz_nb_match_serv = re_serv.GetMatchCount() ;
      wxString s_val ;
      wxString s_dc  ;
      for( size_t sz_num = 2 ; sz_num < co_sz_nb_match_serv ; ++sz_num )
      {
         /*--( Extracting and "cleaning" a value )-------------------------*/
         s_val = re_serv.GetMatch( p_c_data_serv, sz_num ) ;
         re_clean_space.ReplaceAll( &s_val, " " ) ;
         re_clean.ReplaceAll( &s_val, "" ) ;
         s_val = s_val.Trim().Trim( false ) ;

         /*----------------------------------------------------------------*/
         switch( sz_num )
         {
            /*-------------------------------------------------------------*/
            case  2 : server.m_s_price = s_val ; break ;
            case  3 : s_dc = s_val ; break ;
            case  4 : server.m_s_order_url = s_val ; break ;
            case  5 : server.m_s_detail_url = s_val ; break ;
            /*-------------------------------------------------------------*/
            case  7 : case 8 :
               server.m_s_cpu += co_s_col_info_sep ;  // Fallthrough
            case  6 :
               server.m_s_cpu += s_val ;
               break ;
            /*-------------------------------------------------------------*/
            case  9 : server.m_s_memory = s_val ; break ;
            case 10 : server.m_s_storage = s_val ; break ;
            case 11 : server.m_s_network = s_val ; break ;
            /*-------------------------------------------------------------*/
         }
         /*----------------------------------------------------------------*/
      }

      /*--( The same server kind is added for different dc with diff refs )*/
      add_one_server_per_dc( server, s_dc, vec_server_save ) ;
      /*-------------------------------------------------------------------*/
   }
   /*----------------------------------------------------------------------*/
}

/*-------------------------------------------------------------------------*/
void CSourceSoYouStart::refresh_standard( const wxString &co_s_data,
                                          const t_vec_server &vec_server_save
                                        )
{
   /*-----------------------------------------------------------------------+
   ! The So-You-Start server data with similar structures can be divided    !
   ! into two parts :                                                       !
   ! - The "advertised" offers : big individual squares, normally 3 of them !
   !   the tag "div" is used                                                !
   ! - The "listed" offers : One or more "table" per "country"              !
   !   the tag "p" is used                                                  !
   +-----------------------------------------------------------------------*/
   const wxString co_s_re_serv
       = R"=(<p class="prod-name)="
         R"=(.*?>(.*?)<)="                              // Name
         R"=(.*?<div class="servers-.*?>(.*?)<)="       // CPU Brand & Type
         R"=(.*?>(.*?)<)="                              // CPU Cores/Threads
         R"=(.*?>(.*?)</div>)="                         // CPU Speed
         R"=(.*?<div class="servers-.*?>(.*?)</div>)="  // Memory
         R"=(.*?<div class="servers-.*?>(.*?)</div>)="  // Storage
         R"=(.*?<select>(.*?)</select>)="               // Data Center
         R"=(.*?"red".*?>(.*?)</div><div)="             // Price
         R"=(.*?href="(.*?)")="                         // Order url
         R"=(.*?href="(.*?)")="                         // Info url
         ;
   /*----------------------------------------------------------------------*/
   wxRegEx re_serv( co_s_re_serv, wxRE_ADVANCED ) ;
   wxRegEx re_clean_name( R"=(<span .*?</span>|<br>)=", wxRE_ADVANCED ) ;
   wxRegEx re_clean_space( R"=(\s+)=", wxRE_ADVANCED ) ;
   wxRegEx re_clean( R"=(<span[^>]*>|</span>|<br>|<del>.*?</del>)=",
                     wxRE_ADVANCED
                   ) ;
   /*----------------------------------------------------------------------*/
   const wxString co_s_col_info_sep
                = wxGetApp().m_frame->m_notebook->m_co_s_col_info_sep ;
   /*----------------------------------------------------------------------*/
   size_t sz_start_serv ;
   size_t sz_len_serv   ;
   /*----------------------------------------------------------------------*/
   for( const wxChar *p_c_data_serv = co_s_data.c_str() ;
        re_serv.Matches( p_c_data_serv ) ;
        p_c_data_serv += sz_start_serv + sz_len_serv
      )
   {
      /*--( Keep the found element to skip it later )----------------------*/
      re_serv.GetMatch( &sz_start_serv, &sz_len_serv, 0 ) ;
      /*-------------------------------------------------------------------*/
      CServer server ;
      /*--( Clean potential "flash sale" data )----------------------------*/
      server.m_s_name = re_serv.GetMatch( p_c_data_serv, 1 ) ;
      re_clean_name.ReplaceAll( &server.m_s_name, "" ) ;
      /*--( Extract the different parts of the server description )--------*/
      const size_t co_sz_nb_match_serv = re_serv.GetMatchCount() ;
      wxString s_val ;
      wxString s_dc  ;
      for( size_t sz_num = 2 ; sz_num < co_sz_nb_match_serv - 1 ; ++sz_num )
      {
         /*--( Extracting and "cleaning" a value )-------------------------*/
         s_val = re_serv.GetMatch( p_c_data_serv, sz_num ) ;
         re_clean_space.ReplaceAll( &s_val, " " ) ;
         re_clean.ReplaceAll( &s_val, "" ) ;
         s_val = s_val.Trim().Trim( false ) ;

         /*----------------------------------------------------------------*/
         switch( sz_num )
         {
            /*-------------------------------------------------------------*/
            case 3 : case 4 :
               server.m_s_cpu += co_s_col_info_sep ;  // Fallthrough
            case 2 :
               server.m_s_cpu += s_val ;
               break ;
            /*-------------------------------------------------------------*/
            case 5 : server.m_s_memory = s_val ; break ;
            case 6 : server.m_s_storage = s_val ; break ;
            case 7 : s_dc = s_val ; break ;
            case 8 : server.m_s_price = s_val ; break ;
            /*-------------------------------------------------------------*/
         }
         /*----------------------------------------------------------------*/
      }

      /*--( The last elements which don't need any modification )----------*/
      server.m_s_order_url  = re_serv.GetMatch( p_c_data_serv,  9 ) ;
      server.m_s_detail_url = re_serv.GetMatch( p_c_data_serv, 10 ) ;

      /*--( The same server kind is added for different dc with diff refs )*/
      add_one_server_per_dc( server, s_dc, vec_server_save ) ;
      /*-------------------------------------------------------------------*/
   }
   /*----------------------------------------------------------------------*/
}

/*-------------------------------------------------------------------------*/
int CSourceSoYouStart::refresh( int i_err_code, const wxString &co_s_data )
{
   /*----------------------------------------------------------------------*/
   int i_ret_pre = pre_refresh( i_err_code, co_s_data ) ;
   if( i_ret_pre != 0 ) { return post_refresh( i_ret_pre ) ; }

   /*--( Save and reset previous server data )-----------------------------*/
   t_vec_server vec_server_save ;
   m_vec_server.swap( vec_server_save ) ;
   /*----------------------------------------------------------------------*/
   refresh_deals( co_s_data, vec_server_save ) ;
   refresh_standard( co_s_data, vec_server_save ) ;
   /*----------------------------------------------------------------------*/
   return post_refresh( 0 ) ;
   /*----------------------------------------------------------------------*/
}

/*-------------------------------------------------------------------------*/
CNotebook::CNotebook( wxWindow *parent )
         : wxNotebook( parent, wxID_ANY )
{
   /*----------------------------------------------------------------------*/
   m_il = new wxImageList( 16, 16 ) ;
   /*----------------------------------------------------------------------*/
   m_i_img_flag_canada  = m_il->Add( wxIcon( canada_16x16_xpm ) ) ;
   m_i_img_flag_france  = m_il->Add( wxIcon( france_16x16_xpm ) ) ;
   m_i_img_flag_germany = m_il->Add( wxIcon( germany_16x16_xpm ) ) ;
   m_i_img_flag_poland  = m_il->Add( wxIcon( poland_16x16_xpm ) ) ;
     m_i_img_flag_united_kingdom
   = m_il->Add( wxIcon( united_kingdom_16x16_xpm ) ) ;
   /*----------------------------------------------------------------------*/
   for( int i_num = 0 ; i_num < STY_NB ; ++i_num )
   {  /*-------------------------------------------------------------------*/
      const e_server_type co_sty = static_cast< e_server_type >( i_num ) ;
        m_a_i_img_source[ i_num ]
      = m_il->Add( wxIcon( wxGetApp().get_sty_xpm( co_sty ) ) ) ;
      /*-------------------------------------------------------------------*/
   }
   /*----------------------------------------------------------------------*/
   SetImageList( m_il ) ;
   /*----------------------------------------------------------------------*/
   Bind( wxEVT_RIGHT_DOWN,
         [ this ]( wxMouseEvent &event )
         {  context_menu() ; }
       ) ;
   /*----------------------------------------------------------------------*/
   m_timer_check_avail.SetOwner( this ) ;
   Bind( wxEVT_TIMER,
         [ this ]( wxTimerEvent &event )
         {  launch_check_avail( false ) ; },
         m_timer_check_avail.GetId()
       ) ;
   /*-----------------------------------------------------------------------+
   ! Download management                                                    !
   +-----------------------------------------------------------------------*/
   Bind( sc::EVT_DOWNLOAD,
         [ this ]( sc::CDownloadEvent &event )
         {
            /*-------------------------------------------------------------*/
            const int co_i_download_id = event.get_download_id() ;
            /*-------------------------------------------------------------*/
            if(    co_i_download_id >= DID_SERVER_TYPE_START
                && co_i_download_id <= DID_SERVER_TYPE_END
              )
            {  /*----------------------------------------------------------*/
               const int co_i_sty = co_i_download_id - DID_SERVER_TYPE_START;
               /*----------------------------------------------------------*/
               get_server_list( co_i_sty )
                      ->refresh( event.get_err_code(), event.get_string() ) ;
               /*----------------------------------------------------------*/
            }
            else
            if(    co_i_download_id >= DID_AVAIL_START
                && co_i_download_id <= DID_AVAIL_END
              )
            {  /*----------------------------------------------------------*/
               const int co_i_sty = co_i_download_id - DID_AVAIL_START ;
               /*----------------------------------------------------------*/
               get_server_list( co_i_sty )
                              ->check_avail( event.get_err_code(),
                                             event.get_user_data(),
                                             event.get_string()
                                           ) ;
               /*----------------------------------------------------------*/
            }
            else
            {  wxFAIL_MSG( "Unmanaged download id" ) ; }
            /*-------------------------------------------------------------*/
         }
         /*----------------------------------------------------------------*/
      ) ;

   /*----------------------------------------------------------------------*/
}

/*-------------------------------------------------------------------------*/
CNotebook::~CNotebook()
{
   /*----------------------------------------------------------------------*/
   delete m_il ;
   /*----------------------------------------------------------------------*/
}

/*-------------------------------------------------------------------------*/
void CNotebook::add_page_server_list( CServerList *server_list )
{
   /*----------------------------------------------------------------------*/
   AddPage( server_list, server_list->m_s_name, false,
            m_a_i_img_source[ server_list->m_sty ]
          ) ;
   /*----------------------------------------------------------------------*/
}

/*-------------------------------------------------------------------------*/
void CNotebook::add_pages()
{
   /*--( Create the different server sources )-----------------------------*/
   add_page_server_list( new CSourceKimsufi( this ) ) ;
   for( int i_sty = STY_KIMSUFI + 1 ; i_sty < STY_NB ; ++i_sty )
   {
      /*-------------------------------------------------------------------*/
      CServerList *server_list
      = new CSourceSoYouStart( this, static_cast< e_server_type >( i_sty ) );
      /*-------------------------------------------------------------------*/
      add_page_server_list( server_list ) ;
      /*-------------------------------------------------------------------*/
   }
   /*--( Restore width of columns to what they were during last run )------*/
   string_to_col_width( wxGetApp().M_s_serv_list_col_width.get() ) ;
   /*----------------------------------------------------------------------*/
}

/*-------------------------------------------------------------------------*/
int CNotebook::set_select( const wxString &s_ref, bool boo_select )
{
   /*----------------------------------------------------------------------*/
   const int co_i_nb = GetPageCount() ;
   for( int i_num = 0 ; i_num < co_i_nb ; ++i_num )
   {
      /*-------------------------------------------------------------------*/
      CServerList *server_list = get_server_list( i_num ) ;
      CServer     *p_server    = server_list->find_server( s_ref ) ;
      /*--( Server found ? Ref is supposed to be unique among all types )--*/
      if( p_server != nullptr )
      {  long l_item = p_server - server_list->get_server( 0 ) ;
         return server_list->set_select( l_item, boo_select ) ;
      }
      /*-------------------------------------------------------------------*/
   }
   /*--( Not found )-------------------------------------------------------*/
   return -1 ;
   /*----------------------------------------------------------------------*/
}

/*-------------------------------------------------------------------------*/
void CNotebook::select_ref_to_select()
{
   /*----------------------------------------------------------------------*/
   if( m_vec_ref_to_select.empty() ) { return ; }
   /*----------------------------------------------------------------------*/
   for( const auto &s_ref : m_vec_ref_to_select )
   {  set_select( s_ref, true ) ; }
   /*----------------------------------------------------------------------*/
   m_vec_ref_to_select.clear() ;
   /*----------------------------------------------------------------------*/
}

/*-------------------------------------------------------------------------*/
void CNotebook::string_to_col_width( const wxString &s_val )
{
   /*----------------------------------------------------------------------*/
   const int co_i_nb_page = GetPageCount() ;
   wxStringTokenizer tkz_page( s_val, '\n' ) ;
   /*----------------------------------------------------------------------*/
   for( int i_page = 0 ;
        i_page < co_i_nb_page && tkz_page.HasMoreTokens() ;
        ++i_page
      )
   {
      /*-------------------------------------------------------------------*/
      CServerList *server_list = get_server_list( i_page ) ;
      wxStringTokenizer tkz_width( tkz_page.GetNextToken(), '\t' ) ;
      /*-------------------------------------------------------------------*/
      for( int i_col = 0 ;
           i_col < CServerList::COL_NB && tkz_width.HasMoreTokens() ;
           ++i_col
         )
      {  /*----------------------------------------------------------------*/
         int i_width = wxAtoi( tkz_width.GetNextToken() ) ;
         /*--( Minimal check )---------------------------------------------*/
         if( i_width >= 10 )
         {  server_list->SetColumnWidth( i_col, i_width ) ; }
         /*----------------------------------------------------------------*/
      }
      /*-------------------------------------------------------------------*/
   }
   /*----------------------------------------------------------------------*/
}

/*-------------------------------------------------------------------------*/
wxString CNotebook::col_width_to_string() const
{
   /*----------------------------------------------------------------------*/
   wxString s_res ;
   /*----------------------------------------------------------------------*/
   const int co_i_nb_page = GetPageCount() ;
   /*----------------------------------------------------------------------*/
   for( int i_page = 0 ; i_page < co_i_nb_page ; ++i_page )
   {
      /*-------------------------------------------------------------------*/
      if( i_page > 0 && !s_res.empty() ) { s_res += '\n' ; }
      /*-------------------------------------------------------------------*/
      CServerList *server_list = get_server_list( i_page ) ;
      /*-------------------------------------------------------------------*/
      for( int i_col = 0 ; i_col < CServerList::COL_NB ; ++i_col )
      {
         /*----------------------------------------------------------------*/
         if( i_col > 0 && !s_res.empty() ) { s_res += '\t' ; }
         /*----------------------------------------------------------------*/
         s_res += wxString::Format( "%d",
                                    server_list->GetColumnWidth( i_col )
                                  ) ;
         /*----------------------------------------------------------------*/
      }
      /*-------------------------------------------------------------------*/
   }
   /*----------------------------------------------------------------------*/
   return s_res ;
   /*----------------------------------------------------------------------*/
}

/*-------------------------------------------------------------------------*/
void CNotebook::string_to_serv_ref_sel( const wxString &s_val )
{
   /*----------------------------------------------------------------------*/
   wxStringTokenizer tkz( s_val, '\n' ) ;
   wxString s_token ;
   /*----------------------------------------------------------------------*/
   while( tkz.HasMoreTokens() )
   {  m_vec_ref_to_select.emplace_back( tkz.GetNextToken() ) ; }
   /*----------------------------------------------------------------------*/
}

/*-------------------------------------------------------------------------*/
wxString CNotebook::serv_ref_sel_to_string() const
{
   /*----------------------------------------------------------------------*/
   wxString s_res ;
   const int co_i_nb = GetPageCount() ;
   /*----------------------------------------------------------------------*/
   for( int i_num = 0 ; i_num < co_i_nb ; ++i_num )
   {
      /*-------------------------------------------------------------------*/
      CServerList *server_list = get_server_list( i_num ) ;
      /*-------------------------------------------------------------------*/
      for( const auto &server : server_list->m_vec_server )
      {  /*----------------------------------------------------------------*/
         if( !server.m_boo_select ) { continue ; }
         /*----------------------------------------------------------------*/
         if( !s_res.empty() ) { s_res += '\n' ; }
         s_res += server.m_s_ref ;
         /*----------------------------------------------------------------*/
      }
      /*-------------------------------------------------------------------*/
   }
   /*----------------------------------------------------------------------*/
   return s_res ;
   /*----------------------------------------------------------------------*/
}

/*--( Starting the download of the server web pages )----------------------*/
void CNotebook::launch_refresh_server()
{
   /*----------------------------------------------------------------------*/
   CFrame *frame = wxGetApp().m_frame ;
   frame->gui_enable_refresh_server( false ) ;
   frame->gui_enable_check_avail( false ) ;
   /*----------------------------------------------------------------------*/
   frame->log_message( wxString::Format( _( "Refresh server list #%d (%s)"),
                                         ++m_i_nb_refresh_server,
                                         wxGetApp().get_region_name()
                                       )
                     ) ;
   /*----------------------------------------------------------------------*/
   m_i_source_refresh_done = 0 ;
   /*----------------------------------------------------------------------*/
   const int co_i_nb = GetPageCount() ;
   for( int i_num = 0 ; i_num < co_i_nb ; ++i_num )
   {  get_server_list( i_num )->launch_refresh() ; }
   /*----------------------------------------------------------------------*/
}

/*-------------------------------------------------------------------------*/
int CNotebook::launch_check_avail( bool boo_only_current_page )
{
   /*----------------------------------------------------------------------*/
   const wxString co_s_title
                = wxString::Format( _( "Refresh availabilities #%d (%s)" ),
                                    ++m_i_nb_check_avail,
                                    wxGetApp().get_region_name()
                                  ) ;
   /*----------------------------------------------------------------------*/
   CFrame *frame = wxGetApp().m_frame ;
   frame->gui_enable_refresh_server( false ) ;
   frame->gui_enable_check_avail( false ) ;
   /*----------------------------------------------------------------------*/
   m_i_avail_refresh_todo = 0 ;
   m_i_avail_refresh_done = 0 ;
   m_i_page_to_show_avail_ok = -1 ;
   m_l_item_to_show_avail_ok = -1 ;
   /*----------------------------------------------------------------------*/
   const int co_i_nb      = GetPageCount() ;
   const int co_i_current = GetSelection() ;
   /*----------------------------------------------------------------------*/
   for( int i_num = 0 ; i_num < co_i_nb ; ++i_num )
   {
      /*--( Check availabilities for current page servers ? )--------------*/
      if( boo_only_current_page && i_num != co_i_current ) { continue ; }

      /*-------------------------------------------------------------------*/
      CServerList *server_list = get_server_list( i_num ) ;
      const CServer *co_p_server_first = server_list->get_server( 0 ) ;
      /*--( One download per selected server )-----------------------------*/
      for( auto &server : server_list->m_vec_server )
      {
         /*--( Reset current server availability if any )------------------*/
         if( server.m_avail != CServer::AVA_UNKNOWN )
         {  server.m_avail = CServer::AVA_UNKNOWN ;
            server_list->RefreshItem( &server - co_p_server_first ) ;
         }

         /*----------------------------------------------------------------*/
         if( !server.m_boo_select ) { continue ; }
         /*----------------------------------------------------------------*/
         sc::CDownload dl( this, server.get_avail_url() ) ;
         dl.get_dlbase().set_user_data( server.m_s_ref ) ;
         dl.get_dlbase().set_download_id(   DID_AVAIL_START
                                          + server_list->m_sty
                                        ) ;
         dl.thread_start() ;
         ++m_i_avail_refresh_todo ;
         /*----------------------------------------------------------------*/
      }
      /*-------------------------------------------------------------------*/
   }
   /*----------------------------------------------------------------------*/
   if( m_i_avail_refresh_todo == 0 )
   {
      /*-------------------------------------------------------------------*/
      frame->log_error( wxString::Format( _( "%s: no server selected" ),
                                          co_s_title
                                        )
                      ) ;
      /*-------------------------------------------------------------------*/
      frame->gui_enable_refresh_server( true ) ;
      frame->gui_enable_check_avail( true ) ;
      /*-------------------------------------------------------------------*/
      return -1 ;
      /*-------------------------------------------------------------------*/
   }
   /*----------------------------------------------------------------------*/
   frame->log_message( wxString::Format( wxPLURAL( _( "%s: %d server" ),
                                                   _( "%s: %d servers" ),
                                                   m_i_avail_refresh_todo
                                                 ),
                                         co_s_title, m_i_avail_refresh_todo
                                       )
                     ) ;
   /*----------------------------------------------------------------------*/
   return 0 ;
   /*----------------------------------------------------------------------*/
}

/*-------------------------------------------------------------------------*/
void CNotebook::show_info_after_avail_ok()
{
   /*----------------------------------------------------------------------*/
   if( m_i_page_to_show_avail_ok >= 0 )
   {  SetSelection( m_i_page_to_show_avail_ok ) ; }
   /*----------------------------------------------------------------------*/
   if( m_l_item_to_show_avail_ok >= 0 )
   {  get_current_server_list()->EnsureVisible( m_l_item_to_show_avail_ok );}
   /*----------------------------------------------------------------------*/
}

/*-------------------------------------------------------------------------*/
int CNotebook::start_auto_check_avail()
{
   /*----------------------------------------------------------------------*/
   if( get_auto_check_avail_running() ) { return -1 ; }
   /*----------------------------------------------------------------------*/
   CFrame *frame = wxGetApp().m_frame ;
   /*----------------------------------------------------------------------*/
   const int co_i_sec = wxGetApp().M_i_check_avail_period.get() ;
   m_timer_check_avail.Start( co_i_sec * 1000 ) ;
   frame->log_message( wxString::Format(
                    _( "Availability: automatic check done every %ds (%s)" ),
                    co_i_sec, sc::sec2hms( co_i_sec )
                                       )
                     ) ;
   /*----------------------------------------------------------------------*/
   frame->gui_init_auto_check_avail() ;
   /*----------------------------------------------------------------------*/
   return 0 ;
   /*----------------------------------------------------------------------*/
}

/*-------------------------------------------------------------------------*/
int CNotebook::stop_auto_check_avail()
{
   /*----------------------------------------------------------------------*/
   CFrame *frame = wxGetApp().m_frame ;
   /*----------------------------------------------------------------------*/
   m_timer_check_avail.Stop() ;
   frame->log_message( _( "Availability: automatic check stopped" ) ) ;
   /*----------------------------------------------------------------------*/
   frame->gui_init_auto_check_avail() ;
   /*----------------------------------------------------------------------*/
   return 0 ;
   /*----------------------------------------------------------------------*/
}

/*-------------------------------------------------------------------------*/
void CNotebook::context_menu()
{
   /*--( Tab on which the click has been done )----------------------------*/
   int i_sty = HitTest( ScreenToClient( wxGetMousePosition() ) ) ;
   if( i_sty == wxNOT_FOUND ) { return ; }
   /*----------------------------------------------------------------------*/
   wxMenu menu ;
   int i_id = 0 ;
   const int co_i_id_open_url = i_id++ ;
   const int co_i_id_copy_url = i_id   ;
   /*----------------------------------------------------------------------*/
   sc::append_item( &menu, co_i_id_open_url,
                    _( "Open page in &browser" ),
                    sc::get_menu_bmp( wxART_EXECUTABLE_FILE )
                  ) ;
   sc::append_item( &menu, co_i_id_copy_url,
                    _( "Copy &page address" ),
                    sc::get_menu_bmp( wxART_COPY )
                  ) ;
   /*----------------------------------------------------------------------*/
   int i_sel_id = GetPopupMenuSelectionFromUser( menu ) ;
   if( i_sel_id == co_i_id_open_url )
   {  wxLaunchDefaultBrowser( get_server_list( i_sty )->get_main_url() ) ; }
   else
   if( i_sel_id == co_i_id_copy_url )
   {  sc::set_clipboard( get_server_list( i_sty )->get_main_url() ) ; }
   /*----------------------------------------------------------------------*/
}

/*-------------------------------------------------------------------------*/



/*==========================================================================+
!                       End of file CServerList.cpp                         !
+==========================================================================*/
